/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2008 EBM WebSourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.tools.webadmin.datacollector.client;

import java.util.HashMap;
import java.util.Map;

import javax.security.auth.login.FailedLoginException;

import org.ow2.petals.jmx.PetalsJMXClient;
import org.ow2.petals.jmx.exception.ConnectionErrorException;
import org.ow2.petals.tools.webadmin.datacollector.client.exception.DataCollectorClientException;
import org.ow2.petals.tools.webadmin.datacollector.client.service.AdminService;
import org.ow2.petals.tools.webadmin.datacollector.client.service.DeployService;
import org.ow2.petals.tools.webadmin.datacollector.client.service.EndpointService;
import org.ow2.petals.tools.webadmin.datacollector.client.service.InstallService;
import org.ow2.petals.tools.webadmin.datacollector.client.service.MonitoringService;
import org.ow2.petals.tools.webadmin.datacollector.client.service.SystemInfoService;
import org.ow2.petals.tools.webadmin.datacollector.client.service.TransportMonitoringService;
import org.ow2.petals.jmx.api.IPetalsAdminService.ContainerKeyNames;

/**
 * @author mcarpentier - EBM WebSourcing
 * @author cgirodengo - Capgemini Sud
 * 
 *  A {@link CollectorClient} allows you to connect to a petals server through
 *  JMX. A {@link CollectorClient} should be built for each petals server to
 *  admin
 * 
 */
public class CollectorClient {

    private final static int CONNECTION_FAIL_LIMIT = 3;

    private PetalsJMXClient petalsJMXClient;

    private String host;

    private Integer jmxJNDIPort;

    private String jmxLogin;

    private String jmxPassword;

    /** collector services * */
    private MonitoringService monitoringService;

    private AdminService adminService;

    private DeployService deployService;

    private InstallService installService;

    private SystemInfoService systemInfoService;

    private TransportMonitoringService transportMonitoringService;

    private EndpointService endpointService;

    // attributes of associated petals container
    private Map<String, String> containerProperties;

    /** for connection error check* */
    private int connectionFailedCount = 0;

    public CollectorClient(final String host, final Integer jmxJNDIPort, final String jmxLogin,
            final String jmxPassword) {
        this.host = host;
        this.jmxJNDIPort = jmxJNDIPort;
        this.jmxLogin = jmxLogin;
        this.jmxPassword = jmxPassword;
    }

    public CollectorClient(Map<String, String> containerProperties) {
        this.containerProperties = new HashMap<String, String>(containerProperties);

        this.host = this.containerProperties.get(ContainerKeyNames.CONF_HOST);
        this.jmxJNDIPort = Integer.valueOf(this.containerProperties
                .get(ContainerKeyNames.CONF_JMX_JNDI_PORT));
        this.jmxLogin = this.containerProperties.get(ContainerKeyNames.CONF_USER);
        this.jmxPassword = this.containerProperties.get(ContainerKeyNames.CONF_PWD);
    }

    public void connect() throws DataCollectorClientException {
        // if it is a reconnection, reset the fields
        if (this.petalsJMXClient != null) {
            this.reset();
        }

        try {
            this.petalsJMXClient = new PetalsJMXClient(host, jmxJNDIPort, jmxLogin, jmxPassword);

        } catch (final ConnectionErrorException e) {
            this.connectionFailedCount++;
            String errMsg = "Can't connect to PEtALS server on " + host + ":" + jmxJNDIPort;
            if ((e.getCause() != null) && (e.getCause().getCause() != null)
                    && (e.getCause().getCause() instanceof FailedLoginException)) {
                errMsg = errMsg + " - Invalid username or password";
            }
            throw new DataCollectorClientException(errMsg, e);

        }
    }

    public void disconnect() throws DataCollectorClientException {
        try {
            if (this.petalsJMXClient != null) {
                this.petalsJMXClient.disconnect();
            }

        } catch (final ConnectionErrorException e) {
            throw new DataCollectorClientException("Can't disconnect to PEtALS server on " + host
                    + ":" + jmxJNDIPort, e);
        }

        this.reset();
    }

    public synchronized AdminService getAdminService() throws DataCollectorClientException {
        if (this.adminService == null) {
            if (this.petalsJMXClient == null) {
                this.connect();
            }
            this.adminService = new AdminService(this.petalsJMXClient);
        }
        return this.adminService;
    }

    public synchronized DeployService getDeployService() throws DataCollectorClientException {
        if (this.deployService == null) {
            if (this.petalsJMXClient == null) {
                this.connect();
            }
            this.deployService = new DeployService(this.petalsJMXClient);
        }
        return this.deployService;
    }

    public synchronized EndpointService getEndpointService() throws DataCollectorClientException {
        if (this.endpointService == null) {
            if (this.petalsJMXClient == null) {
                this.connect();
            }
            this.endpointService = new EndpointService(this.petalsJMXClient);
        }
        return this.endpointService;
    }

    public synchronized InstallService getInstallService() throws DataCollectorClientException {
        if (this.installService == null) {
            if (this.petalsJMXClient == null) {
                this.connect();
            }
            this.installService = new InstallService(this.petalsJMXClient);
        }
        return this.installService;
    }

    public synchronized MonitoringService getMonitoringService()
            throws DataCollectorClientException {
        if (this.monitoringService == null) {
            if (this.petalsJMXClient == null) {
                this.connect();
            }
            this.monitoringService = new MonitoringService(this.petalsJMXClient);
        }
        return this.monitoringService;
    }

    public synchronized SystemInfoService getSystemInfoService()
            throws DataCollectorClientException {
        if (this.systemInfoService == null) {
            if (this.petalsJMXClient == null) {
                this.connect();
            }
            this.systemInfoService = new SystemInfoService(this.petalsJMXClient);
        }
        return this.systemInfoService;
    }

    public synchronized TransportMonitoringService getTransportMonitoringService()
            throws DataCollectorClientException {
        if (this.transportMonitoringService == null) {
            if (this.petalsJMXClient == null) {
                this.connect();
            }
            this.transportMonitoringService = new TransportMonitoringService(this.petalsJMXClient);
        }
        return this.transportMonitoringService;
    }

    public Map<String, String> getContainerProperties() {
        return containerProperties;
    }

    public void addContainerProperty(final String property, final String value) {
        if (this.containerProperties == null) {
            this.containerProperties = new HashMap<String, String>();
        }
        this.containerProperties.put(property, value);
    }

    public String getServerProperty(String property) {
        if (this.containerProperties == null) {
            this.containerProperties = new HashMap<String, String>();
        }
        return this.containerProperties.get(property);
    }

    /**
     * This method check if ther's not too many connection failed AND inc
     * connection fail count
     * 
     * @return false if connection count has reach the limit
     */
    public boolean checkConnectionFailedCount() {
        return (this.connectionFailedCount < CONNECTION_FAIL_LIMIT);
    }

    public void resetConnectionFailedCount() {
        this.connectionFailedCount = 0;
    }

    public int getConnectionFailedCount() {
        return connectionFailedCount;
    }

    private void reset() {
        this.petalsJMXClient = null;
        this.adminService = null;
        this.deployService = null;
        this.endpointService = null;
        this.installService = null;
        this.monitoringService = null;
        this.systemInfoService = null;
        this.transportMonitoringService = null;
    }

}
