/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2008 EBM WebSourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: CollectorServer.java mcarpentier $
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.tools.webadmin.datacollector.server;

import java.sql.Connection;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.apache.log4j.Logger;
import org.hibernate.HibernateException;
import org.hibernate.Session;
import org.hsqldb.Server;
import org.ow2.petals.jmx.api.IPetalsAdminService.ContainerKeyNames;
import org.ow2.petals.jmx.api.IPetalsAdminService.ContainerStates;
import org.ow2.petals.jmx.api.IPetalsAdminService.DomainKeyNames;
import org.ow2.petals.jmx.api.IPetalsAdminService.SubDomainKeyNames;
import org.ow2.petals.jmx.api.IPetalsAdminService.TopologyKeyNames;
import org.ow2.petals.jmx.exception.ConnectionErrorException;
import org.ow2.petals.tools.webadmin.datacollector.client.CollectorClient;
import org.ow2.petals.tools.webadmin.datacollector.client.exception.DataCollectorClientException;
import org.ow2.petals.tools.webadmin.datacollector.client.exception.DataCollectorClientInitializationException;
import org.ow2.petals.tools.webadmin.datacollector.dao.DAOException;
import org.ow2.petals.tools.webadmin.datacollector.server.exception.DataCollectorClientDoesNotExistException;
import org.ow2.petals.tools.webadmin.datacollector.server.exception.DataCollectorServerException;
import org.ow2.petals.tools.webadmin.datacollector.server.exception.DataCollectorServerNotInitializedException;
import org.ow2.petals.tools.webadmin.datacollector.util.CollectorThread;
import org.ow2.petals.tools.webadmin.datacollector.util.HibernateHelper;
import org.ow2.petals.tools.webadmin.datacollector.util.Queuesmanager;
import org.ow2.petals.tools.webadmin.datacollector.util.ReconnectionThread;
import org.ow2.petals.tools.webadmin.datacollector.util.StringHelper;

public final class CollectorServer {

    private Map<String, CollectorClient> collectorClients;

    private Set<Map<String, String>> subdomainsProperties;

    private Map<String, String> domainProperties;

    private static Server hsqlServer;

    private CollectorThread collectorThread;

    private int queuesCollectingInterval = 10000;

    private int reconnectionInterval = 10000;

    private int reconnectionAttempts = 30;

    private long startTime;

    private ReconnectionThread reconnectionThread;

    private Queuesmanager qm;

    final static Logger log = Logger.getLogger(CollectorServer.class);

    /** properties given to the collector for initialization * */
    private String host;

    private String jndiPort;

    private String user;

    private String password;

    /** properties given to the collector for database initialization * */
    public final static String dbName = "datacollector-database";

    public final static String dbDirectory = "datacollector-hsqldb";

    /**
     * 
     * @throws DataCollectorClientException
     * @throws DataCollectorClientInitializationException
     * @throws DataCollectorServerException
     *      SubDomainKeyNames
     */
    public CollectorServer(String host, String jndiPort, String user, String pwd)
            throws DataCollectorServerException {
        this.qm = new Queuesmanager();

        this.host = host;
        this.jndiPort = jndiPort;
        this.password = pwd;
        this.user = pwd;
        this.startTime = (new Date()).getTime();

        this.collectorClients = new HashMap<String, CollectorClient>();
        this.subdomainsProperties = new HashSet<Map<String, String>>();

        initServer(false);
    }

    /**
     * Do all server init operations.<br>
     * First temp client will try to connect to petals and the retrieve all
     * connected PEtALS servers conf
     * 
     * @param firstClient
     * @throws DataCollectorServerException
     */
    public void initServer(boolean refresh) throws DataCollectorServerException {

        if (refresh) {
            for (CollectorClient collectorClient : this.collectorClients.values()) {
                try {
                    if (collectorClient.getMonitoringService().isMonitoring()) {
                        collectorClient.getMonitoringService().stopListening();
                    }
                } catch (DataCollectorClientException e) {
                    log.error(e);
                }
                try {
                    collectorClient.disconnect();
                } catch (DataCollectorClientException e) {
                    log.error(e);
                }

            }
            this.collectorClients.clear();
            this.subdomainsProperties.clear();
        }

        try {
            final CollectorClient firstClient = new CollectorClient(this.host, Integer
                    .valueOf(this.jndiPort), this.user, this.password);
            final Set<Map<String, String>> topology = firstClient.getAdminService().getTopology();
            firstClient.disconnect();
            for (Map<String, String> topologyElement : topology) {
                if (DomainKeyNames.TOPOLOGY_TYPE.equals(topologyElement
                        .get(TopologyKeyNames.TOPOLOGY_TYPE))) {
                    domainProperties = topologyElement;
                } else if (SubDomainKeyNames.TOPOLOGY_TYPE.equals(topologyElement
                        .get(TopologyKeyNames.TOPOLOGY_TYPE))) {
                    subdomainsProperties.add(topologyElement);
                } else if (ContainerKeyNames.TOPOLOGY_TYPE.equals(topologyElement
                        .get(TopologyKeyNames.TOPOLOGY_TYPE))) {

                    // add only the started Container
                    if (ContainerStates.STATE_STARTED.equals(topologyElement
                            .get(ContainerKeyNames.CONF_STATE))) {
                        // switch the server view JMX RMI port to client
                        // view JMX JNDI port
                        topologyElement.put(ContainerKeyNames.CONF_JMX_JNDI_PORT, topologyElement
                                .remove(ContainerKeyNames.CONF_JMX_RMI_PORT));

                        CollectorClient cc = new CollectorClient(topologyElement);

                        this.collectorClients.put(cc
                                .getServerProperty(ContainerKeyNames.CONF_CONTAINER_NAME), cc);
                        try {
                            // If monitoring for this server is activated, start
                            // listening for messages
                            if (cc.getMonitoringService().isMonitoring()) {
                                cc.getMonitoringService().startListening();
                            }

                        } catch (DataCollectorClientException e) {
                            // The monitoring of this server is not
                            // available
                            log.debug(e);
                            cc.addContainerProperty(ContainerKeyNames.CONF_STATE,
                                    ContainerStates.STATE_STOPPED);
                        }
                    }
                }
            }

            // queues listening
            if (!refresh) {
                this.startQueuesListening();
            }

        } catch (final DataCollectorClientException e) {
            throw new DataCollectorServerException(e);
        } catch (NumberFormatException e) {
            throw new DataCollectorServerException("Can't init JMX connection", e);
        }
    }

    /**
     * Check JDBC connection
     * 
     * @throws DataCollectorServerException
     */
    @SuppressWarnings("deprecation")
    public void checkDatabaseIsRunning() throws DataCollectorServerException {
        Session session = null;
        try {
            session = HibernateHelper.getSession();
            final Connection con = session.connection();
            log.debug("Connection to database server OK : " + con.toString());
        } catch (final HibernateException e) {
            if (session != null) {
                session.close();
            }
            throw new DataCollectorServerException("Can't open connection to database server", e);
        }

    }

    public CollectorClient getCollectorClient(String serverName)
            throws DataCollectorServerException, DataCollectorClientDoesNotExistException,
            DataCollectorServerNotInitializedException {
        if ((this.collectorClients == null) || (this.collectorClients.isEmpty())) {
            throw new DataCollectorServerNotInitializedException();
        } else {
            if (this.collectorClients.containsKey(serverName)) {
                return this.collectorClients.get(serverName);
            } else {
                throw new DataCollectorClientDoesNotExistException("Can't find " + serverName
                        + " JMX client");
            }
        }
    }

    public Map<String, String> getDomainProperties()
            throws DataCollectorServerNotInitializedException {
        if (this.domainProperties.isEmpty()) {
            throw new DataCollectorServerNotInitializedException();
        }

        return this.domainProperties;
    }

    public Set<Map<String, String>> getSubdomainsProperties()
            throws DataCollectorServerNotInitializedException {
        if (this.subdomainsProperties.isEmpty()) {
            throw new DataCollectorServerNotInitializedException();
        }

        return this.subdomainsProperties;
    }

    public Set<Map<String, String>> getContainersProperties()
            throws DataCollectorServerNotInitializedException {
        if (this.collectorClients.isEmpty()) {
            throw new DataCollectorServerNotInitializedException();
        } else {
            final Set<Map<String, String>> result = new HashSet<Map<String, String>>();
            for (final String client : this.collectorClients.keySet()) {
                result.add(this.collectorClients.get(client).getContainerProperties());
            }
            return result;
        }
    }

    public CollectorClient getDefaultCollectorClient() throws DataCollectorServerException,
            DataCollectorServerNotInitializedException {
        CollectorClient defaultClient = null;
        if ((this.collectorClients == null) || (this.collectorClients.isEmpty())) {
            throw new DataCollectorServerNotInitializedException();
        } else {
            defaultClient = this.collectorClients.values().iterator().next();
        }
        return defaultClient;

    }

    public static void startDatabaseServer(int port) throws DataCollectorServerException {
        try {
            if (hsqlServer == null) {
                hsqlServer = (Server) Class.forName("org.hsqldb.Server").newInstance();
                hsqlServer.setDatabasePath(0, dbDirectory + "/" + dbName);
                hsqlServer.setDatabaseName(0, dbName);
                hsqlServer.setPort(port);
                hsqlServer.setSilent(true);
                hsqlServer.setTrace(false);
            }
            hsqlServer.checkRunning(true);
        } catch (final RuntimeException re) {
            // server is not running, start it
            hsqlServer.start();
        } catch (InstantiationException e) {
            throw new DataCollectorServerException("Can't start database server", e);
        } catch (IllegalAccessException e) {
            throw new DataCollectorServerException("Can't start database server", e);
        } catch (ClassNotFoundException e) {
            throw new DataCollectorServerException("Can't start database server", e);

        }
        String url = "jdbc:hsqldb:hsql://localhost:" + port + "/" + dbName;
        HibernateHelper.configureSession("org.hsqldb.jdbcDriver", url, "sa", "");

    }

    public void startQueuesListening() {
        this.stopQueuesListening();
        // TODO : implement this in a proper way
        this.collectorThread = new CollectorThread(this);
        this.collectorThread.start();
    }

    public static void stopDatabaseServer() {
        if (hsqlServer != null) {
            Session session = null;
            try {
                session = HibernateHelper.getSession();
                // SQLQuery q = session.createSQLQuery("SHUTDOWN IMMEDIATELY");
                hsqlServer.stop();
                hsqlServer.shutdown();
            } catch (final Exception e) {
                log.error("Can't shutdown database", e);
            } finally {
                if (session != null)
                    session.close();
            }

        }
    }

    public void stopQueuesListening() {
        if ((this.collectorThread != null) && (this.collectorThread.isAlive())) {
            this.collectorThread.interrupt();
        }
    }

    public void stopMessagesListening() {
        try {
            if (this.collectorClients != null) {
                for (CollectorClient c : this.collectorClients.values()) {
                    c.getMonitoringService().stopListening();
                }
            }
        } catch (DataCollectorClientException e) {
            /** we just need to log the error, no more* */
            log.info("Couldn't remove message listener from PEtALS", e);
        }
    }

    public void collectAllQueues() {
        String currentClient = "";
        try {
            if (this.collectorClients != null) {
                for (String client : this.collectorClients.keySet()) {
                    currentClient = client;
                    if (qm != null) {
                        qm.collectQueues(client, this.collectorClients.get(client)
                                .getTransportMonitoringService());
                    }
                }
            }
        } catch (DataCollectorClientException e) {
            CollectorServer.log.error("Can't retrieve queues status", e);
            if (e.getCause() instanceof ConnectionErrorException) {
                /** check client connection timeout* */

                if ((StringHelper.isNotNullAndNotEmpty(currentClient))
                        && (this.collectorClients.containsKey(currentClient))) {
                    CollectorClient cc = this.collectorClients.get(currentClient);
                    if (!cc.checkConnectionFailedCount()) {
                        this.collectorClients.remove(currentClient);
                        if ((this.collectorClients == null) || (this.collectorClients.isEmpty())) {
                            log
                                    .info("PEtALS connection seems to be lost, reconnection timer starting");
                            this.reconnectionThread = new ReconnectionThread(this);
                            this.reconnectionThread.start();
                        }
                    }
                }
            }
        } catch (DAOException e) {
            CollectorServer.log.error("Can't retreive queues status", e);
        }

    }

    public int getQueuesCollectingInterval() {
        return this.queuesCollectingInterval;
    }

    public void setQueuesCollectingInterval(int freq) {
        this.queuesCollectingInterval = freq;
        if (this.collectorThread != null) {
            this.collectorThread.setRefreshFrequency(freq);
        }
    }

    public int getReconnectionInterval() {
        return reconnectionInterval;
    }

    public void setReconnectionInterval(int reconnectionInterval) {
        this.reconnectionInterval = reconnectionInterval;
        if (this.reconnectionThread != null) {
            this.reconnectionThread.setRefreshFrequency(reconnectionInterval);
        }
    }

    public int getReconnectionAttempts() {
        return reconnectionAttempts;
    }

    public void setReconnectionAttempts(int reconnectionAttempts) {
        this.reconnectionAttempts = reconnectionAttempts;
        if (this.reconnectionThread != null) {
            this.reconnectionThread.setMaxAttempts(reconnectionAttempts);
        }
    }

    public String getDatabasePath() {
        return hsqlServer.getDatabasePath(0, true);
    }

    public int getDatabasePort() {
        return hsqlServer.getPort();
    }

    public long getStartTime() {
        return startTime;
    }

    public String getDatabaseVersion() {
        String result = "unknown";
        if (CollectorServer.hsqlServer != null) {
            result = CollectorServer.hsqlServer.getProductName() + " "
                    + CollectorServer.hsqlServer.getProductVersion();
        }
        return result;
    }

}
